import abc
from pathlib import Path
from typing import Optional

from ..typing import TypePath
from ..transforms import Transform
from ..download import download_and_extract_archive
from .. import SubjectsDataset, Subject, ScalarImage, LabelMap


class BITE(SubjectsDataset, abc.ABC):
    base_url = 'http://www.bic.mni.mcgill.ca/uploads/Services/'

    def __init__(
            self,
            root: TypePath,
            transform: Optional[Transform] = None,
            download: bool = False,
            **kwargs,
            ):
        root = Path(root).expanduser().absolute()
        if download:
            self._download(root)
        subjects_list = self._get_subjects_list(root)
        self.kwargs = kwargs
        super().__init__(subjects_list, transform=transform, **kwargs)

    def _download(self, root: Path):
        raise NotImplementedError

    def _get_subjects_list(self, root: Path):
        raise NotImplementedError


class BITE3(BITE):
    dirname = 'group3'
    """Pre- and post-resection MR images in BITE.

    *The goal of BITE is to share in vivo medical images of patients wtith
    brain tumors to facilitate the development and validation of new image
    processing algorithms.*

    Please check the `BITE website`_ for more information and
    acknowledgments instructions.

    .. _BITE website: http://nist.mni.mcgill.ca/?page_id=672

    Args:
        root: Root directory to which the dataset will be downloaded.
        transform: An instance of
            :class:`~torchio.transforms.transform.Transform`.
        download: If set to ``True``, will download the data into :attr:`root`.
    """
    def _download(self, root: Path):
        if (root / self.dirname).is_dir():
            return
        root.mkdir(exist_ok=True, parents=True)
        filename = f'{self.dirname}.tar.gz'
        url = self.base_url + filename
        download_and_extract_archive(
            url,
            download_root=root,
            md5='e415b63887c40b727c45552614b44634',
        )
        (root / filename).unlink()  # cleanup

    def _get_subjects_list(self, root: Path):
        subjects_dir = root / self.dirname
        subjects = []
        for i in range(1, 15):
            if i == 13:
                continue  # no MRI for this subject
            subject_id = f'{i:02d}'
            subject_dir = subjects_dir / subject_id
            preop_path = subject_dir / f'{subject_id}_preop_mri.mnc'
            postop_path = subject_dir / f'{subject_id}_postop_mri.mnc'
            images_dict = {}
            images_dict['preop'] = ScalarImage(preop_path)
            images_dict['postop'] = ScalarImage(postop_path)
            for fp in subject_dir.glob('*tumor*'):
                images_dict[fp.stem[3:]] = LabelMap(fp)
            subject = Subject(images_dict)
            subjects.append(subject)
        return subjects



import torch
import numpy as np
from torchio.transforms import HistogramStandardization
from torchio import LabelMap, ScalarImage, Subject, SubjectsDataset
from ...utils import TorchioTestCase


class TestHistogramStandardization(TorchioTestCase):
    """Tests for :class:`HistogramStandardization` class."""

    def setUp(self):
        super().setUp()
        subjects = []
        for i in range(5):
            image = ScalarImage(self.get_image_path(f'hs_image_{i}'))
            label_path = self.get_image_path(
                f'hs_label_{i}', binary=True, force_binary_foreground=True)
            label = LabelMap(label_path)
            subject = Subject(image=image, label=label)
            subjects.append(subject)
        self.subjects = subjects
        self.dataset = SubjectsDataset(self.subjects)

    def test_train_histogram(self):
        paths = [subject.image.path for subject in self.dataset]
        # Use a function to mask
        HistogramStandardization.train(
            paths,
            masking_function=HistogramStandardization.mean,
            output_path=(self.dir / 'landmarks.txt'),
        )
        # Use a file to mask
        HistogramStandardization.train(
            paths,
            mask_path=self.dataset[0].label.path,
            output_path=(self.dir / 'landmarks.npy'),
        )
        # Use files to mask
        masks = [subject.label.path for subject in self.dataset]
        HistogramStandardization.train(
            paths,
            mask_path=masks,
            output_path=(self.dir / 'landmarks_masks.npy'),
        )

    def test_bad_paths_lengths(self):
        with self.assertRaises(ValueError):
            HistogramStandardization.train(
                [1, 2],
                mask_path=[1, 2, 3],
            )

    def test_normalize(self):
        landmarks = np.linspace(0, 100, 13)
        landmarks_dict = {'image': landmarks}
        transform = HistogramStandardization(landmarks_dict)
        transform(self.dataset[0])

    def test_wrong_image_key(self):
        landmarks = np.linspace(0, 100, 13)
        landmarks_dict = {'wrong_key': landmarks}
        transform = HistogramStandardization(landmarks_dict)
        with self.assertRaises(KeyError):
            transform(self.dataset[0])

    def test_with_saved_dict(self):
        landmarks = np.linspace(0, 100, 13)
        landmarks_dict = {'image': landmarks}
        torch.save(landmarks_dict, self.dir / 'landmarks_dict.pth')
        landmarks_dict = torch.load(self.dir / 'landmarks_dict.pth')
        transform = HistogramStandardization(landmarks_dict)
        transform(self.dataset[0])

    def test_with_saved_array(self):
        landmarks = np.linspace(0, 100, 13)
        np.save(self.dir / 'landmarks.npy', landmarks)
        landmarks_dict = {'image': self.dir / 'landmarks.npy'}
        transform = HistogramStandardization(landmarks_dict)
        transform(self.dataset[0])
